import SubscriptionController from "./SubscriptionController";
import EventEmitter from "wolfy87-eventemitter";

class PouchDBSubscription extends EventEmitter {
	constructor(db, {selector = null, findSelector = null, subscribeSelector = null, find = false}) {
		super();
		this._db = db;
		this._subscriptionControllers = new Set();
		this._canceled = false;

		this._dbChangeOptions = {};

		if(find) { // try use selector
			this._dbChangeOptions.find = true;
			if (typeof findSelector === 'object' && findSelector !== null)
			    	this._dbChangeOptions.findSelector = findSelector;
			else if (typeof selector === 'object' && selector !== null)
				this._dbChangeOptions.findSelector = selector;

			if (typeof subscribeSelector === 'object' && subscribeSelector !== null)
				this._dbChangeOptions.subscribeSelector = subscribeSelector;
			else if (typeof selector === 'object' && selector !== null)
				this._dbChangeOptions.subscribeSelector = selector;
		}
	}

	static get MUTATIONTYPES() {
		return Object.freeze({
			UPSERT: 'UPSERT',
			DELETE: 'DELETE'
		});
	}

	/**
	 * seq of last change recorded
	 * @returns {null}
	 */
	get seq() {
		return this.hasOwnProperty('_dbSeq') ? this._dbSeq : null;
	}

	get canceled() {
		return this._canceled;
	}

	get subscriptions() {
		return this._subscriptionControllers.size;
	}

	cancel() {
		if(this.canceled)
			throw new Error('already_canceled');
		else {
			this._canceled = true;

			if(this.hasOwnProperty('_dbChanges')) {
				this._dbChanges.cancel();
				delete this._dbChanges;
				delete this._dbSeq;
			}
			for(let subscriptionController of this._subscriptionControllers)
				subscriptionController.cancel();
			this._subscriptionControllers.empty();

			this.emit('cancel');
		}
	}

	async listenForChanges() {
		const self = this;

		if(!self.hasOwnProperty('_dbChanges')) {
			const dbChanges = await self._db.changes({since: 'now'});
			self._dbSeq = dbChanges.last_seq;

			let changeOptions = {
				live: true,
				since: 'now',
				include_docs: true,
			};

			if(this._dbChangeOptions.find) {
				changeOptions.find = true;
				changeOptions.selector = this._dbChangeOptions.subscribeSelector;
			}

			self._dbChanges = this._db.changes(changeOptions)
				.on('change', (dbChange) => {
					self._dbSeq = dbChange.seq;
					if(dbChange.deleted)
						self._mutation(PouchDBSubscription.MUTATIONTYPES.DELETE, {_id: dbChange.id}, dbChange.seq);
					else
						self._mutation(PouchDBSubscription.MUTATIONTYPES.UPSERT, dbChange.doc, dbChange.seq);
				});
		}
	}

	_mutation(mutationType, doc, seq) {
		this.emit('mutation', mutationType, doc, seq);
	}

	subscribe(since = 'now') {
		const self = this;

		return new Promise(async (resolve) => {
			const subscriptionController = new SubscriptionController();
			const queuedMutations = new Set();

			await self.listenForChanges();

			self._subscriptionControllers.add(subscriptionController);
			self.emit('subscriptions', self.subscriptions);

			subscriptionController.once('cancel', () => {
				self.off('mutation', emitOrQueueMutation);
				self._subscriptionControllers.delete(subscriptionController);
				self.emit('subscriptions', self.subscriptions);
			});

			self.on('mutation', emitOrQueueMutation);

			resolve(subscriptionController);

			let findMaxSeq = null;

			if(since === 'now')
				findComplete();
			else if(since === null || Number(since) === 0) { // Since the beginning
				if (typeof self._dbChangeOptions.findSelector === 'object') { // has a selector
					findMaxSeq = self.seq;
					const result = await self._db.find({selector: self._dbChangeOptions.findSelector});
					for (let resultDoc of result.docs)
						subscriptionController.mutation(PouchDBSubscription.MUTATIONTYPES.UPSERT, resultDoc);

					findComplete()
				}
				else { // no selector
					const result = await self._db.allDocs({include_docs: true, update_seq: true});
					findMaxSeq = result.update_seq;
					if(Array.isArray(result.rows)) {
						for(let resultRow of result.rows)
							subscriptionController.mutation(PouchDBSubscription.MUTATIONTYPES.UPSERT, resultRow.doc);
					}
					findComplete();
				}
			}
			else { // since is a number, > 0
				let modifiedDocs = new Set();
				findMaxSeq = self.seq;
				const dbChanges = await self._db.changes({
					live: false,
					include_docs: false,
					since: Number(since)
				});

				// TODO use bulkDocs
				for(let dbChange of dbChanges.results) {
					if(dbChange.seq <= findMaxSeq && !modifiedDocs.has(dbChange.id)) {
						try {
							let resultDoc = await self._db.get(dbChange.id);
							subscriptionController.mutation(PouchDBSubscription.MUTATIONTYPES.UPSERT, resultDoc);
						}
						catch (error) {
							if (error.name === 'not_found')
								subscriptionController.mutation(PouchDBSubscription.MUTATIONTYPES.DELETE, {_id: dbChange.id});
						}
						finally {
							modifiedDocs.add(dbChange.id);
						}
					}
				}
				modifiedDocs.clear(); // cleanup
				findComplete();
			}

			function findComplete() {
				subscriptionController.findComplete = true;
				for(let {mutationType, doc, seq} of queuedMutations)
					subscriptionController.mutation(mutationType, doc, seq);
			}

			function emitOrQueueMutation(mutationType, doc, seq) {
				if(since === 'now' || seq >= since) {
					if(subscriptionController.findComplete)
						subscriptionController.mutation(mutationType, doc, seq);
					else if(findMaxSeq !== null && seq > findMaxSeq)
						queuedMutations.add({mutationType: mutationType, doc: doc, seq: seq});
				}
			}
		});
	}
}

export default PouchDBSubscription;
